import logging as lg
import shutil
from pathlib import Path
import subprocess

from easysam.generate import generate
from easysam.commondep import commondep
import easysam.utils as u

SAM_CLI_VERSION = '1.138.0'
PIP_VERSION = '25.1.1'


def deploy(cliparams, directory, stack):
    resources = generate(directory, [], False)

    lg.info(f'Deploying SAM template from {directory}')
    check_pip_version(cliparams)
    check_sam_cli_version(cliparams)
    remove_common_dependencies(directory)
    copy_common_dependencies(directory, resources)
    sam_build(cliparams, directory)
    sam_deploy(cliparams, directory, stack)

    if not cliparams.get('no_cleanup'):
        remove_common_dependencies(directory)


def delete(cliparams, stack, force):
    lg.info(f'Deleting SAM template from {stack}')
    cf = u.get_aws_client('cloudformation', cliparams)
    mode = 'FORCE_DELETE_STACK' if force else 'STANDARD'
    cf.delete_stack(StackName=stack, DeletionMode=mode)  # type: ignore


def check_pip_version(cliparams):
    lg.info('Checking pip version')

    try:
        lg.debug(f'Running command: {" ".join(["pip", "--version"])}')
        pip_version = subprocess.check_output(['pip', '--version']).decode('utf-8')

        if pip_version < PIP_VERSION:
            raise UserWarning(f'pip version must be {PIP_VERSION} or higher')

    except Exception as e:
        raise UserWarning('pip not found') from e


def check_sam_cli_version(cliparams):
    lg.info('Checking SAM CLI version')
    sam_tool = cliparams['sam_tool']
    sam_params = sam_tool.split(' ')
    sam_params.append('--version')

    try:
        lg.debug(f'Running command: {" ".join(sam_params)}')
        sam_version = subprocess.check_output(sam_params).decode('utf-8')
        lg.debug(f'SAM CLI version: {sam_version}')

        if sam_version < SAM_CLI_VERSION:
            raise UserWarning(f'SAM CLI version must be {SAM_CLI_VERSION} or higher')

    except Exception as e:
        raise UserWarning(f'SAM CLI not found. Error: {e}') from e


def sam_build(cliparams, directory):
    lg.info(f'Building SAM template from {directory}')
    sam_tool = cliparams['sam_tool']
    sam_params = sam_tool.split(' ')
    sam_params.append('build')

    if cliparams.get('verbose'):
        sam_params.append('--debug')

    if cliparams['dry_run']:
        lg.info(f'Would run: {" ".join(sam_params)}')
        return

    lg.debug(f'Running command: {" ".join(sam_params)}')
    subprocess.run(sam_params, cwd=directory.resolve(), text=True, check=True)
    lg.info('Successfully built SAM template')


def sam_deploy(cliparams, directory, aws_stack):
    lg.info(f'Deploying SAM template from {directory} to {aws_stack}')
    sam_tool = cliparams['sam_tool']
    sam_params = sam_tool.split(' ')

    sam_params.extend([
        'deploy',
        '--parameter-overrides', f'ParameterKey=Stage,ParameterValue={aws_stack}',
        '--stack-name', aws_stack,
        '--no-fail-on-empty-changeset',
        '--no-confirm-changeset',
        '--resolve-s3',
        '--capabilities', 'CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'
    ])

    aws_tags = cliparams.get('tag')
    lg.info(f'Using AWS tags: {aws_tags}')
    aws_tag_string = ' '.join(aws_tags)
    lg.debug(f'AWS tag string: {aws_tag_string}')
    sam_params.extend(['--tags', aws_tag_string])

    if region := cliparams.get('region'):
        sam_params.extend(['--region', region])

    if cliparams.get('verbose'):
        sam_params.append('--debug')

    if cliparams['dry_run']:
        lg.info(f'Would run: {" ".join(sam_params)}')
        return

    if cliparams.get('aws_profile'):
        lg.info(f'Using AWS profile: {cliparams["aws_profile"]}')
        sam_params.extend(['--profile', cliparams['aws_profile']])

    lg.debug(f'Running command: {" ".join(sam_params)}')
    subprocess.run(sam_params, cwd=directory.resolve(), text=True, check=True)
    lg.info('Successfully deployed SAM template')


def common_dep_dir(directory):
    return Path(directory, 'common')


def remove_common_dependencies(directory):
    lg.info(f'Removing common dependencies from {directory}')
    backend = Path(directory, 'backend')

    for common_dep in backend.glob('**/common'):
        lg.debug(f'Removing {common_dep}')
        shutil.rmtree(common_dep)


def copy_common_dependencies(directory, resources):
    lg.info(f'Copying common dependencies to {directory}')
    common = common_dep_dir(directory)

    if 'functions' not in resources:
        lg.warning('No functions found in resources')
        return

    for lambda_name, lambda_function in resources['functions'].items():
        lambda_path = Path(directory, lambda_function['uri'])
        lambda_common_path = Path(lambda_path, 'common')
        lambda_common_path.mkdir(parents=True, exist_ok=True)
        deps = commondep(common, lambda_path)

        lg.info(f'Lambda {lambda_name} has {len(deps)} common dependencies')
        lg.debug(f'Dependencies: {" ".join(deps)}')

        for dep in deps:
            dep_path = Path(common, dep)

            if dep_path.is_dir():
                lg.debug(f'Copying {dep_path} directory to {lambda_common_path}')
                lambda_common_dep_path = Path(lambda_common_path, dep_path.name)
                shutil.copytree(dep_path, lambda_common_dep_path)
            else:
                dep_filepath = dep_path.with_suffix('.py')
                lg.debug(f'Copying {dep_filepath} file to {lambda_common_path}')
                lambda_common_filepath = Path(lambda_common_path, dep_filepath.name)
                shutil.copy(dep_filepath, lambda_common_filepath)



# SAM Template Generator and Swagger support for API Gateway

from pathlib import Path
import logging as lg

from jinja2 import Environment, FileSystemLoader
import yaml
import click

import prismarine.prisma_common as prisma_common
import prismarine.prisma_easysam as prisma_easysam

from easysam.prismarine import generate as generate_prismarine_clients


IMPORT_FILE = 'easysam.yaml'


def write_result(path, text):
    path.parent.mkdir(parents=True, exist_ok=True)
    sane_text = '\n'.join(line for line in text.splitlines() if line and line.strip())
    Path(path).write_text(sane_text)


def prismarine_dynamo_tables(prefix, base, package, resources_dir, path):
    lg.debug(f'Generating prismarine dynamo tables for {prefix}')

    try:
        prisma_common.set_path(path)
        base_dir = Path(resources_dir, base).resolve()
        cluster = prisma_common.get_cluster(base_dir, package)
        return prisma_easysam.build_dynamo_tables(prefix, cluster)
    except Exception as e:
        lg.error(f'Error generating dynamo tables for {prefix}: {e}')
        raise UserWarning(f'Error generating prismarine dynamo tables for {base_dir}')


def preprocess_prismarine(resources_data, resources_dir, path):
    prefix = resources_data['prefix']
    prisma = resources_data['prismarine']
    prisma_base = prisma.get('default-base')
    prisma_tables = prisma['tables'] or []

    for prisma_integration in prisma_tables:
        base = prisma_integration.get('base') or prisma_base
        package = prisma_integration.get('package')

        if not base:
            raise UserWarning(f'No base found for {package}')

        if not package:
            raise UserWarning(f'No package found for {base}')

        tables = prismarine_dynamo_tables(prefix, base, package, resources_dir, path)

        if 'tables' not in resources_data:
            resources_data['tables'] = {}

        resources_data['tables'].update(tables)


def preprocess_lambda(resources_data, resources_dir, lambda_def, entry_path, entry_dir):
    if 'functions' not in resources_data:
        resources_data['functions'] = {}

    lambda_name = lambda_def.get('name')

    if not lambda_name:
        raise UserWarning(f'Import file {entry_path} contains no lambda name')

    if lambda_name in resources_data['functions']:
        raise UserWarning(
            f'Import file {entry_path} contains duplicate lambda name {lambda_name}'
        )

    lambda_resources = lambda_def.get('resources', {})

    if 'uri' not in lambda_resources:
        lambda_resources['uri'] = Path(entry_dir).relative_to(resources_dir).as_posix()

    lg.debug(f'Adding lambda {lambda_name} to resources')
    resources_data['functions'][lambda_name] = lambda_resources
    integration = lambda_def.get('integration', {})

    if integration:
        path = integration.get('path')

        if not path:
            raise UserWarning(f'Import file {entry_path} contains no path name')

        if 'paths' not in resources_data:
            resources_data['paths'] = {}

        if path in resources_data['paths']:
            raise UserWarning(f'Import file {entry_path} contains duplicate path {path}')

        integration['function'] = lambda_name
        del integration['path']
        lg.debug(f'Adding path {path} to resources')
        resources_data['paths'][path] = integration


def preprocess_tables(resources_data: dict, table_def: dict, entry_path: Path):
    if 'tables' not in resources_data:
        resources_data['tables'] = {}

    for table_name, table_data in table_def.items():
        if table_name in resources_data['tables']:
            raise UserWarning(f'Import file {entry_path} contains duplicate table {table_name}')

        lg.debug(f'Adding table {table_name} to resources')
        resources_data['tables'][table_name] = table_data


def preprocess_file(resources_data: dict, resources_dir: Path, entry_path: Path):
    entry_dir = entry_path.parent
    entry_data = yaml.safe_load(entry_path.read_text())
    lg.info(f'Processing import file {entry_path}')

    if not all(key in ['lambda', 'import', 'tables'] for key in entry_data.keys()):
        raise UserWarning(f'Import file {entry_path} contains unexpected sections')

    if lambda_def := entry_data.get('lambda'):
        preprocess_lambda(resources_data, resources_dir, lambda_def, entry_path, entry_dir)

    if tables_def := entry_data.get('tables'):
        preprocess_tables(resources_data, tables_def, entry_path)

    if local_import_def := entry_data.get('import'):
        for import_file in local_import_def:
            import_path = Path(entry_dir, import_file)
            preprocess_file(resources_data, resources_dir, import_path)


def preprocess_imports(resources_data: dict, resources_dir: Path):
    for import_dir_str in resources_data.get('import', []):
        import_dir = Path(resources_dir, import_dir_str)
        lg.info(f'Processing import directory {import_dir}')

        if not import_dir.exists():
            raise UserWarning(f'Import directory {import_dir} not found')

        for entry_path in import_dir.glob(f'**/{IMPORT_FILE}'):
            preprocess_file(resources_data, resources_dir, entry_path)


def preprocess_resources(resources_data, resources_dir, path):
    def sort_dict(d):
        return dict(sorted(d.items(), key=lambda x: x[0]))

    if 'prismarine' in resources_data:
        preprocess_prismarine(resources_data, resources_dir, path)

    if 'import' in resources_data:
        preprocess_imports(resources_data, resources_dir)

    for section in ['tables', 'paths', 'functions', 'buckets', 'authorizers']:
        if section in resources_data:
            resources_data[section] = sort_dict(resources_data[section])

    resources_data = sort_dict(resources_data)


def generate(directory, path, preprocess_only):
    resources_dir = Path(directory)
    path = [resources_dir] + list(path)
    resources = Path(resources_dir, 'resources.yaml')
    build_dir = Path(resources_dir, 'build')
    swagger = Path(build_dir, 'swagger.yaml')
    template = Path(resources_dir, 'template.yml')
    resources_data = yaml.safe_load(Path(resources).read_text())

    lg.info('Processing resources')
    preprocess_resources(resources_data, resources_dir, path)

    if preprocess_only:
        click.echo(yaml.dump(resources_data, indent=4))
        return

    lg.debug('Resources processed:\n' + yaml.dump(resources_data, indent=4))

    loader = FileSystemLoader(searchpath=[
        str(Path(__file__).parent.resolve()),
        str(resources_dir.resolve()),
    ])

    jenv = Environment(loader=loader)
    swagger_template = jenv.get_template('swagger.j2')
    sam_template = jenv.get_template('template.j2')
    swagger_output = swagger_template.render(resources_data)
    sam_output = sam_template.render(resources_data)
    write_result(swagger, swagger_output)
    lg.info(f'Swagger file generated: {swagger}')
    write_result(template, sam_output)
    lg.info(f'SAM template generated: {template}')

    if 'prismarine' in resources_data:
        lg.info('Generating prismarine clients')
        generate_prismarine_clients(resources_dir, resources_data)

    return resources_data



from pathlib import Path

import click

from easysam.commondep import commondep


@click.group(help='Inspect the application for debugging purposes')
def inspect():
    pass


@inspect.command(name='common-deps', help='Inspect a lambda function')
@click.option(
    '--common-dir', type=str, default='common',
    help='The directory containing the common dependencies'
)
@click.argument(
    'lambda-dir', type=click.Path(exists=True)
)
def common_deps(common_dir, lambda_dir):
    common_dir = Path(common_dir)
    lambda_dir = Path(lambda_dir)

    deps = commondep(common_dir, lambda_dir)
    click.echo('Dependencies:')

    for dep in deps:
        click.echo(f'* {dep}')



from pathlib import Path
import logging as lg


UTILS_PY = '''\
# Common code

def my_common_function():
    return 'Hello, world!'
'''

RESOURCES_YAML = '''\
prefix: MyApp

import:
  - backend
'''

LAMBDA_EASY_SAM = '''\
lambda:
  name: myfunction
  resources:
    tables:
    - MyItem
  integration:
    path: /items
    open: true
    greedy: false
'''

INDEX_PY = '''\
# Lambda function code

import json

import common.utils as u


def handler(event, context):
    data = u.my_common_function()

    return {
        'statusCode': 200,
        'headers': {
            'Access-Control-Allow-Headers': '*',
            'Access-Control-Allow-Origin': '*',
            'Access-Control-Allow-Methods': '*'
        },
        'body': json.dumps({'data': data})
    }
'''


DATABASE_EASY_SAM = '''\
tables:
  MyItem:
    attributes:
    - hash: true
      name: ItemID
'''

ROOT_GITIGNORE = '''\
build
template.yml
.aws-sam
'''

FUNCTION_GITIGNORE = '''\
__pycache__/
*.py[oc]
**/common/
'''

REQUIREMENTS_TXT = '''\
boto3
'''


def init(cliparams, app_name):
    lg.info(f'Initializing app {app_name}')
    app_dir = Path(app_name)
    lg.debug(f'Creating app directory {app_dir}')
    app_dir.mkdir(parents=True, exist_ok=True)
    lg.debug(f'Creating resources EasySAM file {app_dir / "resources.yaml"}')
    app_dir.joinpath('resources.yaml').write_text(RESOURCES_YAML)
    lg.debug(f'Creating root .gitignore file {app_dir / ".gitignore"}')
    app_dir.joinpath('.gitignore').write_text(ROOT_GITIGNORE)

    common_dir = app_dir / 'common'
    common_dir.mkdir(parents=True, exist_ok=True)
    lg.debug(f'Creating common directory {common_dir}')
    common_dir.joinpath('utils.py').write_text(UTILS_PY)

    backend_dir = app_dir / 'backend'
    lg.debug(f'Creating backend directory {backend_dir}')
    backend_dir.mkdir(parents=True, exist_ok=True)
    lg.debug(f'Creating database directory {backend_dir}')
    database_dir = backend_dir / 'database'
    database_dir.mkdir(parents=True, exist_ok=True)
    lg.debug(f'Creating database EasySAM file {database_dir / "easysam.yaml"}')
    database_dir.joinpath('easysam.yaml').write_text(DATABASE_EASY_SAM)

    function_dir = backend_dir / 'function'
    lg.debug(f'Creating function directory {function_dir}')
    function_dir.mkdir(parents=True, exist_ok=True)
    lg.debug(f'Creating function .gitignore file {function_dir / ".gitignore"}')
    function_dir.joinpath('.gitignore').write_text(FUNCTION_GITIGNORE)

    lambda_dir = function_dir / 'myfunction'
    lg.debug(f'Creating lambda directory {lambda_dir}')
    lambda_dir.mkdir(parents=True, exist_ok=True)
    lg.debug(f'Creating lambda EasySAM file {lambda_dir / "easysam.yaml"}')
    lambda_dir.joinpath('easysam.yaml').write_text(LAMBDA_EASY_SAM)
    lg.debug(f'Creating lambda index file {lambda_dir / "index.py"}')
    lambda_dir.joinpath('index.py').write_text(INDEX_PY)

    third_party_dir = app_dir / 'thirdparty'
    lg.debug(f'Creating third party directory {third_party_dir}')
    third_party_dir.mkdir(parents=True, exist_ok=True)
    lg.debug(f'Creating third party requirements.txt file {third_party_dir / "requirements.txt"}')
    third_party_dir.joinpath('requirements.txt').write_text(REQUIREMENTS_TXT)



from pathlib import Path
import click
import logging as lg
import sys
from importlib.metadata import version


from easysam.generate import generate
from easysam.deploy import deploy, delete
from easysam.deploy import remove_common_dependencies
from easysam.init import init

from easysam.inspect import inspect


@click.group(help='EasySAM is a tool for generating SAM templates from simple YAML files')
@click.pass_context
@click.option('--aws-profile', type=str, help='AWS profile to use')
@click.option('--verbose', is_flag=True)
def easysam(ctx, verbose, aws_profile):
    ctx.obj = {'verbose': verbose, 'aws_profile': aws_profile}
    lg.basicConfig(level=lg.DEBUG if verbose else lg.INFO)
    lg.debug(f'Verbose: {verbose}')


@easysam.command(name='generate', help='Generate a SAM template from a directory')
@click.option('--path', multiple=True)
@click.option('--preprocess-only', is_flag=True, default=False)
@click.argument('directory', type=click.Path(exists=True))
def generate_cmd(directory, path, preprocess_only):
    generate(directory, path, preprocess_only)


@easysam.command(name='deploy', help='Deploy the application to an AWS stack')
@click.pass_obj
@click.option(
    '--region', type=str, help='AWS region'
)
@click.option(
    '--tag', type=str, multiple=True, help='AWS tags', required=True
)
@click.option(
    '--dry-run', is_flag=True, help='Dry run the deployment'
)
@click.option(
    '--sam-tool', type=str, help='Path to the SAM CLI', default='uv run sam'
)
@click.option(
    '--no-cleanup', is_flag=True, help='Do not clean the directory before deploying'
)
@click.argument('directory', type=click.Path(exists=True))
@click.argument('stack', type=str)
def deploy_cmd(obj, directory, stack, **kwargs):
    obj.update(kwargs)
    directory = Path(directory)
    deploy(obj, directory, stack)


@easysam.command(name='delete', help='Delete the stack from AWS')
@click.pass_obj
@click.option('--force', is_flag=True, help='Force delete the stack')
@click.option('--region', type=str, help='AWS region')
@click.argument('stack', type=str)
def delete_cmd(obj, stack, force, **kwargs):
    obj.update(kwargs)
    delete(obj, stack, force)


@easysam.command(name='cleanup', help='Remove common dependencies from the directory')
@click.pass_obj
@click.argument('directory', type=click.Path(exists=True))
def cleanup_cmd(obj, directory):
    remove_common_dependencies(directory)


@easysam.command(name='init', help='Initialize a new application in a given directory')
@click.pass_obj
@click.argument('app-name', type=click.Path(exists=False))
def init_cmd(obj, app_name):
    init(obj, app_name)


@easysam.command(name='version', help='Print the version of EasySAM')
def version_cmd():
    click.echo(version('easysam'))


def main():
    try:
        easysam.add_command(inspect)
        easysam()

    except UserWarning as e:
        lg.error(e)
        sys.exit(1)

    except Exception as e:
        lg.error(e)
        sys.exit(2)


if __name__ == '__main__':
    main()
